// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:bloc_test/bloc_test.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';
import 'package:wyatt_authentication_bloc/wyatt_authentication_bloc.dart';

class MockAuthenticationRepository extends Mock
    implements AuthenticationRepository {}

class MockUser extends Mock implements User {}

void main() {
  group('AuthenticationCubit<T>', () {
    final MockUser user = MockUser();
    late AuthenticationRepository authenticationRepository;

    setUp(() {
      authenticationRepository = MockAuthenticationRepository();
      when(() => authenticationRepository.user).thenAnswer(
        (_) => const Stream.empty(),
      );
      when(() => authenticationRepository.cubitStatus).thenAnswer(
        (_) => Stream.fromIterable([AuthCubitStatus.stoped]),
      );
      when(
        () => authenticationRepository.currentUser,
      ).thenReturn(user);
    });

    test('initial auth state is `unknown`', () {
      expect(
        AuthenticationCubit<void>(
          authenticationRepository: authenticationRepository,
        ).state,
        const AuthenticationState<Never>.unknown(),
      );
    });

    test('initial cubit status is `stoped`', () async {
      expect(
        await AuthenticationCubit<void>(
          authenticationRepository: authenticationRepository,
        ).status,
        AuthCubitStatus.stoped,
      );
    });

    group('UserChanged', () {
      blocTest<AuthenticationCubit<void>, AuthenticationState<void>>(
        'emits authenticated when user is not empty',
        setUp: () {
          when(() => user.isNotEmpty).thenReturn(true);
          when(() => authenticationRepository.user).thenAnswer(
            (_) => Stream.value(user),
          );
          when(() => authenticationRepository.cubitStatus).thenAnswer(
            (_) => Stream.value(AuthCubitStatus.started),
          );
        },
        build: () => AuthenticationCubit(
          authenticationRepository: authenticationRepository,
        )..init(),
        seed: () => const AuthenticationState.unknown(),
        expect: () => [AuthenticationState<void>.authenticated(user, null)],
      );

      blocTest<AuthenticationCubit<void>, AuthenticationState<void>>(
        'emits unauthenticated when user is empty',
        setUp: () {
          when(() => user.isEmpty).thenReturn(true);
          when(() => user.isNotEmpty).thenReturn(false);
          when(() => authenticationRepository.user).thenAnswer(
            (_) => Stream.value(user),
          );
          when(() => authenticationRepository.cubitStatus).thenAnswer(
            (_) => Stream.value(AuthCubitStatus.started),
          );
        },
        build: () => AuthenticationCubit(
          authenticationRepository: authenticationRepository,
        )..init(),
        seed: () => const AuthenticationState.unknown(),
        expect: () => [const AuthenticationState<Never>.unauthenticated()],
      );
    });

    group('LogoutRequested', () {
      blocTest<AuthenticationCubit<void>, AuthenticationState<void>>(
        'invokes signOut',
        setUp: () {
          when(
            () => authenticationRepository.signOut(),
          ).thenAnswer((_) async {});
        },
        build: () => AuthenticationCubit(
          authenticationRepository: authenticationRepository,
        ),
        act: (cubit) => cubit.logOut(),
        verify: (_) {
          verify(() => authenticationRepository.signOut()).called(1);
        },
      );
    });
  });
}
